@@ -19,11 +19,12 @@
 #ifndef __GST_FFMPEGAUDDEC_H__
 #define __GST_FFMPEGAUDDEC_H__
 
+#include <glib.h>
+
 G_BEGIN_DECLS
 
 #include <gst/gst.h>
 #include <gst/audio/audio.h>
-#include <gst/audio/gstaudiodecoder.h>
 #include <libavcodec/avcodec.h>
 
 typedef struct _GstFFMpegAudDec GstFFMpegAudDec;
@@ -33,19 +34,15 @@ struct _GstFFMpegAudDec
 
   /* decoding */
   AVCodecContext *context;
-  gboolean opened;
 
   AVFrame *frame;
 
   guint8 *padded;
-  guint padded_size;
+  gint padded_size;
 
   /* prevent reopening the decoder on GST_EVENT_CAPS when caps are same as last time. */
   GstCaps *last_caps;
 
-  /* Stores current buffers to push as GstAudioDecoder wants 1:1 mapping for input/output buffers */
-  GstBuffer *outbuf;
-
   /* current output format */
   GstAudioInfo info;
   GstAudioChannelPosition ffmpeg_layout[64];
